@@ -39,11 +39,23 @@
 #include "zoom-region-private.h"
 #include "GNOME_Magnifier.h"
 
-
-#define MAG_WINDOW_OVERRIDE_REDIRECT
-
 static GObjectClass *parent_class = NULL;
 
+enum {
+	STRUT_LEFT = 0,
+	STRUT_RIGHT = 1,
+	STRUT_TOP = 2,
+	STRUT_BOTTOM = 3,
+	STRUT_LEFT_START = 4,
+	STRUT_LEFT_END = 5,
+	STRUT_RIGHT_START = 6,
+	STRUT_RIGHT_END = 7,
+	STRUT_TOP_START = 8,
+	STRUT_TOP_END = 9,
+	STRUT_BOTTOM_START = 10,
+	STRUT_BOTTOM_END = 11
+};
+
 enum {
 	MAGNIFIER_SOURCE_DISPLAY_PROP,
 	MAGNIFIER_TARGET_DISPLAY_PROP,
@@ -63,9 +75,13 @@ enum {
 static int _x_error = 0;
 static int fixes_event_base = 0, fixes_error_base;
 static Display *cursor_client_connection;
+static Magnifier *_this_magnifier = NULL;
 
 static void magnifier_transform_cursor (Magnifier *magnifier);
 static void magnifier_init_cursor_set (Magnifier *magnifier, gchar *cursor_set);
+static gboolean magnifier_check_set_struts (Magnifier *magnifier);
+static gboolean magnifier_reset_struts_at_idle (gpointer data);
+static gboolean _is_override_redirect = FALSE;
 
 gboolean
 magnifier_error_check (void)
@@ -364,7 +380,6 @@ magnifier_expose_filter (GdkXEvent *xevent, GdkEvent *event, gpointer data)
 	}
 	return GDK_FILTER_CONTINUE;
 }
-
 static void
 magnifier_set_extension_listeners (Magnifier *magnifier, GdkWindow *root)
 {
@@ -375,6 +390,12 @@ magnifier_set_extension_listeners (Magnifier *magnifier, GdkWindow *root)
 	magnifier_cursor_notification_init (magnifier);
 }
 
+static void
+magnifier_size_allocate (GtkWidget *widget)
+{
+	magnifier_check_set_struts (_this_magnifier);
+}
+
 static void
 magnifier_realize (GtkWidget *widget)
 {
@@ -781,6 +802,107 @@ magnifier_init_cursor_set (Magnifier *magnifier, gchar *cursor_set)
 	magnifier_transform_cursor (magnifier);
 }
 
+static gboolean 
+magnifier_reset_struts_at_idle (gpointer data)
+{
+	if (data)
+	{
+		Magnifier *magnifier = MAGNIFIER (data);
+		if (magnifier->priv && GTK_WIDGET_REALIZED (magnifier->priv->w) && 
+		    magnifier_check_set_struts (magnifier))
+		{
+			return FALSE;
+		}
+	}
+	return TRUE;
+}
+
+static gboolean
+magnifier_check_set_struts (Magnifier *magnifier)
+{
+	/* TODO: don't do this if we're using Composite */
+	if (magnifier &&
+	    magnifier->priv && magnifier->priv->w && GTK_WIDGET_REALIZED (magnifier->priv->w) &&
+	    magnifier->priv->w->window) 
+	{
+		Atom atom_strut = gdk_x11_get_xatom_by_name ("_NET_WM_STRUT");
+		Atom atom_strut_partial = gdk_x11_get_xatom_by_name ("_NET_WM_STRUT_PARTIAL");
+		guint32 struts[12] = {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0};
+		GtkWidget *widget = magnifier->priv->w;
+		gint width = gdk_screen_get_width (
+			gdk_display_get_screen (magnifier->target_display,
+						magnifier->target_screen_num));
+		gint height = gdk_screen_get_height (
+			gdk_display_get_screen (magnifier->target_display,
+						magnifier->target_screen_num));
+
+		gint right_margin, left_margin, top_margin, bottom_margin;
+		gint wx, wy, ww, wh;
+
+		gtk_window_get_position (GTK_WINDOW (magnifier->priv->w), &wx, &wy);
+		gtk_window_get_size (GTK_WINDOW (magnifier->priv->w), &ww, &wh);
+
+		left_margin = wx;
+		right_margin = (width - ww) - wx;
+		top_margin = wy;
+		bottom_margin = (height - wh) - wy;
+
+		/* set the WM_STRUT properties on the appropriate side */
+		if (bottom_margin > top_margin && 
+		    bottom_margin > left_margin &&
+		    bottom_margin > right_margin)
+		{
+			struts[STRUT_TOP] = wh + wy;
+			struts[STRUT_TOP_START] = wx;
+			struts[STRUT_TOP_END] = wx + ww;
+		} 
+		else if (top_margin > bottom_margin && 
+			 top_margin > left_margin &&
+			 top_margin > right_margin)
+		{
+			struts[STRUT_BOTTOM] = height - wy;
+			struts[STRUT_BOTTOM_START] = wx;
+			struts[STRUT_BOTTOM_END] = wx + ww;
+		}
+		else if (right_margin > left_margin &&
+			 right_margin > top_margin &&
+			 right_margin > bottom_margin)
+		{
+			struts[STRUT_LEFT] = wx;
+			struts[STRUT_LEFT_START] = wy;
+			struts[STRUT_LEFT_END] = wh + wy;
+		}
+		else 
+		{
+			struts[STRUT_RIGHT] = width - wx;
+			struts[STRUT_RIGHT_START] = wy;
+			struts[STRUT_RIGHT_END] = wy + wh;
+		}
+		
+		gdk_error_trap_push ();
+		XChangeProperty (GDK_WINDOW_XDISPLAY (widget->window), 
+				 GDK_WINDOW_XWINDOW (widget->window), 
+				 atom_strut,
+				 XA_CARDINAL, 32, PropModeReplace,
+				 (guchar *) &struts, 4);
+		XChangeProperty (GDK_WINDOW_XDISPLAY (widget->window), 
+				 GDK_WINDOW_XWINDOW (widget->window), 
+				 atom_strut_partial,
+				 XA_CARDINAL, 32, PropModeReplace,
+				 (guchar *) &struts, 12); 
+		gdk_error_trap_pop ();
+
+#ifdef DEBUG_STRUTS
+		g_message ("struts TOP %d (%d - %d)", struts[STRUT_TOP], struts[STRUT_TOP_START], struts[STRUT_TOP_END]);
+		g_message ("struts BOTTOM %d (%d - %d)", struts[STRUT_BOTTOM], struts[STRUT_BOTTOM_START], struts[STRUT_BOTTOM_END]);
+		g_message ("struts LEFT %d (%d - %d)", struts[STRUT_LEFT], struts[STRUT_LEFT_START], struts[STRUT_LEFT_END]);
+		g_message ("struts RIGHT %d (%d - %d)", struts[STRUT_RIGHT], struts[STRUT_RIGHT_START], struts[STRUT_RIGHT_END]);
+#endif
+		return TRUE;
+	}
+	return FALSE;
+}
+
 static void
 magnifier_get_property (BonoboPropertyBag *bag,
 			BonoboArg *arg,
@@ -872,6 +994,7 @@ magnifier_set_property (BonoboPropertyBag *bag,
 					magnifier->source_screen_num));
 		magnifier_warp_cursor_to_screen (magnifier);
 		magnifier_get_display_rect_bounds (magnifier, &magnifier->source_bounds, FALSE);
+		magnifier_check_set_struts (magnifier);
 		break;
 	case MAGNIFIER_TARGET_DISPLAY_PROP:
 		full_display_string = BONOBO_ARG_GET_STRING (arg);
@@ -889,6 +1012,7 @@ magnifier_set_property (BonoboPropertyBag *bag,
 		magnifier_get_display_rect_bounds (magnifier, &magnifier->target_bounds, TRUE);
 		magnifier_get_display_rect_bounds (magnifier, &magnifier->source_bounds, FALSE);
 		magnifier_reparent_zoom_regions (magnifier);
+		magnifier_check_set_struts (magnifier);
 		break;
 	case MAGNIFIER_SOURCE_SIZE_PROP:
 	        magnifier->source_bounds = BONOBO_ARG_GET_GENERAL (arg,
@@ -908,7 +1032,7 @@ magnifier_set_property (BonoboPropertyBag *bag,
 		gtk_window_resize (GTK_WINDOW (magnifier->priv->w),
 				   magnifier->target_bounds.x2 - magnifier->target_bounds.x1,
 				   magnifier->target_bounds.y2 - magnifier->target_bounds.y1);
-		
+		magnifier_check_set_struts (magnifier);
 		break;
 	case MAGNIFIER_CURSOR_SET_PROP:
 		magnifier_init_cursor_set (magnifier, g_strdup (BONOBO_ARG_GET_STRING (arg)));
@@ -1087,6 +1211,7 @@ impl_magnifier_add_zoom_region (PortableServer_Servant servant,
 
 	/* FIXME: this needs proper lifecycle management */
 	magnifier->zoom_regions = g_list_append (magnifier->zoom_regions, region);
+	magnifier_check_set_struts (magnifier);
 
 	return CORBA_TRUE;
 }
@@ -1325,21 +1450,21 @@ magnifier_properties_init (Magnifier *magnifier)
 static void
 magnifier_init_window (Magnifier *magnifier)
 {
+	GtkWindowType mag_win_type = GTK_WINDOW_TOPLEVEL;
+	if (_is_override_redirect) mag_win_type = GTK_WINDOW_POPUP;
+
 	magnifier->priv->w =
 		g_object_connect (gtk_widget_new (gtk_window_get_type (),
 						  "user_data", NULL,
 						  "can_focus", FALSE,
-#ifdef MAG_WINDOW_OVERRIDE_REDIRECT
-						  "type", GTK_WINDOW_POPUP,
-#else
-						  "type", GTK_WINDOW_TOPLEVEL,
-#endif
+						  "type", mag_win_type,  
 						  "title", "magnifier",
 						  "allow_grow", TRUE,
 						  "allow_shrink", TRUE,
 						  "border_width", 0,
 						  NULL),
 				  "signal::realize", magnifier_realize, NULL,
+				  "signal::size_allocate", magnifier_size_allocate, NULL,
 				  "signal::destroy", magnifier_exit, NULL,
 				  NULL);
 	magnifier->priv->canvas = gtk_fixed_new ();
@@ -1397,19 +1522,26 @@ magnifier_get_cursor (Magnifier *magnifier)
 }
 
 Magnifier *
-magnifier_new (void)
+magnifier_new (gboolean override_redirect)
 {
 	Magnifier *mag;
 	int ret;
 
+	_is_override_redirect = override_redirect;
+
 	mag = g_object_new (magnifier_get_type(), NULL);
+
+	_this_magnifier = mag; /* FIXME what about multiple instances? */
+	
 	ret = bonobo_activation_active_server_register (
 		MAGNIFIER_OAFIID, BONOBO_OBJREF (mag));
 	if (ret != Bonobo_ACTIVATION_REG_SUCCESS)
 	    g_error ("Error registering magnifier server.\n");
 	else
 	    g_message ("The magnifier server was successfully registred.\n");
-	
+
+	g_idle_add (magnifier_reset_struts_at_idle, mag);
+
 	return mag;
 }
 
